const utils = require("./utils");
const fetch = require("node-fetch");
const md5 = require("md5");
const moment = require("moment");

const IDENTITY_AUDIENCE_PROD = "api.atlassian.com";
const IDENTITY_AUDIENCE_STAGING = "api.stg.atlassian.com";
const IDENTITY_AUDIENCE_FEDRAMP_PROD = "api.atlassian-us-gov-mod.com";
const IDENTITY_AUDIENCE_FEDRAMP_STAGING = "api.stg.atlassian-us-gov-mod.com";

const TOKEN_ENDPOINT_PROD = "https://auth.atlassian.com/oauth/token";
const TOKEN_ENDPOINT_STAGING = "https://auth.stg.atlassian.com/oauth/token";
const TOKEN_ENDPOINT_FEDRAMP_PROD =
  "https://auth.atlassian-us-gov-mod.com/oauth/token";
const TOKEN_ENDPOINT_FEDRAMP_STAGING =
  "https://auth.stg.atlassian-us-gov-mod.com/oauth/token";

class ForgeOAuth2 {
  constructor(addon) {
    utils.checkNotNull(addon, "addon");
    utils.checkNotNull(addon.settings, "addon.settings");
    this.addon = addon;
  }

  /**
   * hash a data store index key for storing the access token in a cache
   *
   * @param {String} oauthClientId
   * @returns {String} A key which identifies the user's token in the data store
   */
  _hashedTokenCacheStoreKey(oauthClientId) {
    return `bearer:${md5(oauthClientId)}`; // personal information; Hash it.
  }

  /**
   * Looks up a cached bearer token for a given oauthClientId in the data store
   *
   * @param {String} oauthClientId
   * @param {String} clientKey
   * @returns {Promise} A promise that returns the access token if resolved, or an error if rejected
   */
  _getCachedBearerToken(oauthClientId, clientKey) {
    utils.checkNotUndefined(oauthClientId, "oauthClientId");
    utils.checkNotUndefined(clientKey, "clientSettings.clientKey");

    const key = this._hashedTokenCacheStoreKey(oauthClientId);
    return this.addon.settings.get(key, clientKey);
  }

  /**
   * Stores the bearer token in a cache
   *
   * @param {String} oauthClientId - oauthClientId
   * @param {String} bearerToken - bearerToken to be cached
   * @param {String} expiresAt - The time when the token expires
   * @param {String} clientKey
   * @returns {Promise} A promise that is resolved when the key is stored
   */
  _cacheBearerToken(oauthClientId, bearerToken, expiresAt, clientKey) {
    utils.checkNotUndefined(oauthClientId);
    utils.checkNotUndefined(clientKey);

    const key = this._hashedTokenCacheStoreKey(oauthClientId);
    const token = {
      token: bearerToken,
      expiresAt
    };

    return this.addon.settings.set(key, token, clientKey);
  }

  /**
   * Requesting the bearer token to token endpoint (auth0 proxy)
   *
   * @param {Object} clientSettings - Settings object for the current tenant
   * @returns {Promise} A promise that returns the token object if resolved, or an error if rejected
   */
  async _getBearerToken(clientSettings) {
    utils.checkNotUndefined(clientSettings.baseUrl, "clientSettings.baseUrl");
    utils.checkNotUndefined(
      clientSettings.oauthClientId,
      "clientSettings.oauthClientId"
    );
    utils.checkNotUndefined(
      clientSettings.sharedSecret,
      "clientSettings.sharedSecret"
    );

    const tokenEndpoint = this.getTokenEndpoint(clientSettings.baseUrl);
    const identityAudience = this.getIdentityAudience(clientSettings.baseUrl);

    const payload = {
      grant_type: "client_credentials",
      client_id: clientSettings.oauthClientId,
      client_secret: clientSettings.sharedSecret,
      audience: identityAudience
    };

    try {
      const response = await fetch(tokenEndpoint, {
        method: "POST",
        body: JSON.stringify(payload),
        headers: {
          "Content-Type": "application/json"
        }
      });

      return await response.json();
    } catch (error) {
      throw new Error(
        `HTTP error while getting the bearer token from: ${error}`
      );
    }
  }

  getIdentityAudience(hostBaseUrl) {
    let identityAudience = IDENTITY_AUDIENCE_PROD;
    if (utils.isProductionHost(hostBaseUrl)) {
      identityAudience = IDENTITY_AUDIENCE_PROD;
    } else if (utils.isFedRAMPProductionHost(hostBaseUrl)) {
      identityAudience = IDENTITY_AUDIENCE_FEDRAMP_PROD;
    } else if (utils.isJiraDevBaseUrl(hostBaseUrl)) {
      identityAudience = IDENTITY_AUDIENCE_STAGING;
    } else if (utils.isFedRAMPStagingHost(hostBaseUrl)) {
      identityAudience = IDENTITY_AUDIENCE_FEDRAMP_STAGING;
    }

    return identityAudience;
  }

  getTokenEndpoint(hostBaseUrl) {
    let tokenEndpoint = TOKEN_ENDPOINT_PROD;
    if (utils.isProductionHost(hostBaseUrl)) {
      tokenEndpoint = TOKEN_ENDPOINT_PROD;
    } else if (utils.isFedRAMPProductionHost(hostBaseUrl)) {
      tokenEndpoint = TOKEN_ENDPOINT_FEDRAMP_PROD;
    } else if (utils.isJiraDevBaseUrl(hostBaseUrl)) {
      tokenEndpoint = TOKEN_ENDPOINT_STAGING;
    } else if (utils.isFedRAMPStagingHost(hostBaseUrl)) {
      tokenEndpoint = TOKEN_ENDPOINT_FEDRAMP_STAGING;
    }

    return tokenEndpoint;
  }

  /**
   * Requesting the bearer token considering the cache lifetime
   *
   * @param {Object} clientSettings - Settings object for the current tenant
   * @returns {Promise} A promise that returns the token object if resolved, or an error if rejected
   */
  async getBearerToken(clientSettings) {
    try {
      const cachedToken = await this._getCachedBearerToken(
        clientSettings.oauthClientId,
        clientSettings.clientKey
      );

      if (cachedToken) {
        // cut the expiry time by a few seconds for leeway
        const tokenExpiryTime = moment
          .unix(cachedToken.expiresAt)
          .subtract(3, "seconds");
        const isTokenExpired = tokenExpiryTime.isBefore(moment());
        if (!isTokenExpired) {
          return cachedToken.token;
        }
      }

      // no available cache -> need to re-request a token from the server
      const now = moment();
      const token = await this._getBearerToken(clientSettings);

      // reset the cache
      const tokenExpiry = now.add(token.expires_in, "seconds").unix();
      await this._cacheBearerToken(
        clientSettings.oauthClientId,
        token,
        tokenExpiry,
        clientSettings.clientKey
      );
      return token;
    } catch (error) {
      throw new Error(`error while getting the bearer token: ${error}`);
    }
  }
}

module.exports = ForgeOAuth2;
